from __future__ import annotations

import asyncio
import os
import time

import aiofiles
import aiohttp
import requests
import tenacity
from result import Err, Ok, Result

from .. import messages as m
from .. import t


def isOk(x: t.Any) -> t.TypeGuard[Ok]:
    return isinstance(x, Ok)


def isErr(x: t.Any) -> t.TypeGuard[Err]:
    return isinstance(x, Err)


ghPrefix = "https://raw.githubusercontent.com/speced/bikeshed-boilerplate/main/"


def update(path: str, dryRun: bool = False) -> set[str] | None:
    try:
        m.say("Downloading boilerplates...")
        data = requests.get(ghPrefix + "manifest.txt", timeout=5).text
    except Exception as e:
        m.die(f"Couldn't download boilerplates manifest.\n{e}")
        return None

    newPaths: list[str] = pathsFromManifest(data)
    goodPaths: list[str] = []
    badPaths: list[str] = []

    if not dryRun:
        m.say(
            f"Updating {len(newPaths)} file{'s' if len(newPaths) > 1 else ''}...",
        )
        goodPaths, badPaths = asyncio.run(updateFiles(path, newPaths))
    if not badPaths:
        m.say("Done!")
        return set(goodPaths)
    else:
        phrase = f"were {len(badPaths)} errors" if len(badPaths) > 1 else "was 1 error"
        m.die(
            f"Done, but there {phrase} (of {len(newPaths)} total) in downloading or saving. Run `bikeshed update` again to retry.",
        )
        return set(goodPaths)


def pathsFromManifest(manifest: str) -> list[str]:
    lines = manifest.split("\n")[1:]
    return [line.partition(" ")[2] for line in lines if line != ""]


async def updateFiles(localPrefix: str, newPaths: t.Sequence[str]) -> tuple[list[str], list[str]]:
    tasks = set()
    async with aiohttp.ClientSession(trust_env=True) as session:
        for filePath in newPaths:
            coro = updateFile(localPrefix, filePath, session=session)
            tasks.add(coro)

        lastMsgTime = time.time()
        messageDelta = 2
        goodPaths = []
        badPaths = []
        for future in asyncio.as_completed(tasks):
            result = await future
            if result.is_ok():
                goodPaths.append(result.value)
            else:
                badPaths.append(result.value)
            currFileTime = time.time()
            if (currFileTime - lastMsgTime) >= messageDelta:
                if not badPaths:
                    m.say(f"Updated {len(goodPaths)}/{len(newPaths)}...")
                else:
                    m.say(f"Updated {len(goodPaths)}/{len(newPaths)}, {len(badPaths)} errors...")
                lastMsgTime = currFileTime
    return goodPaths, badPaths


async def updateFile(localPrefix: str, filePath: str, session: t.Any) -> Result[str, str]:
    remotePath = ghPrefix + filePath
    localPath = localizePath(localPrefix, filePath)
    res = await downloadFile(remotePath, session)
    if isOk(res):
        res = await saveFile(localPath, res.ok())
    else:
        m.warn(f"Error downloading {filePath}, full error was:\n{await errorFromAsyncErr(res)}")
    ret: Result[str, str]
    if isErr(res):
        ret = Err(filePath)
    else:
        ret = t.cast("Ok[str]", res)
    return ret


async def errorFromAsyncErr(res: Result[str, t.Awaitable[str]]) -> str | Exception:
    if isOk(res):
        return t.cast(str, res.ok())
    try:
        x = await t.cast("t.Awaitable[str]", res.err())
    except Exception as e:
        return e
    return x


def wrapError(retry_state: t.Any) -> Err[t.Awaitable[str]]:
    return Err(asyncio.wrap_future(retry_state.outcome))


@tenacity.retry(
    reraise=True,
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_random(1, 2),
    retry_error_callback=wrapError,
)
async def downloadFile(path: str, session: t.Any) -> Result[str, t.Awaitable[str]]:
    resp = await session.request(method="GET", url=path)
    resp.raise_for_status()
    return Ok(await resp.text())


@tenacity.retry(
    reraise=True,
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_random(1, 2),
    retry_error_callback=wrapError,
)
async def saveFile(path: str, data: str) -> Result[str, t.Awaitable[str]]:
    dirPath = os.path.dirname(path)
    if not os.path.exists(dirPath):
        os.makedirs(dirPath)
    async with aiofiles.open(path, "w", encoding="utf-8") as fh:
        await fh.write(data)
        return Ok(path)


def localizePath(root: str, relPath: str) -> str:
    return os.path.join(root, *relPath.split("/"))



from . import result
from .main import (
    debugNodes,
    initialDocumentParse,
    linesFromNodes,
    nodesFromHtml,
    nodesFromStream,
    parseLines,
    parseText,
    parseTitle,
    strFromNodes,
)
from .nodes import (
    Comment,
    Doctype,
    EndTag,
    ParserNode,
    RawElement,
    RawText,
    SafeText,
    SelfClosedTag,
    StartTag,
    Text,
)
from .stream import (
    ParseConfig,
    Stream,
)



from __future__ import annotations

import hashlib
import itertools
import re

from . import datablocks, h, markdown, t
from . import messages as m


def processInclusions(doc: t.SpecT) -> None:
    iters = 0
    while True:
        # Loop because an include can contain more includes.
        iters += 1
        if iters > 1000:
            m.die("Looked for include-blocks more than 1000 times, something is wrong.")
            return
        els = h.findAll("pre.include", doc)
        if not els:
            break
        for el in els:
            handleBikeshedInclude(el, doc)
    for el in h.findAll("pre.include-code", doc):
        handleCodeInclude(el, doc)
    for el in h.findAll("pre.include-raw", doc):
        handleRawInclude(el, doc)


def handleBikeshedInclude(el: t.ElementT, doc: t.SpecT) -> None:
    macros = {}
    for i in itertools.count(0):
        macroLine = el.get("macro-" + str(i))
        if macroLine is None:
            break
        k, _, v = macroLine.partition(" ")
        macros[k.lower()] = v
    if el.get("path"):
        path = el.get("path")
        assert path is not None
        includedInputSource = doc.inputSource.relative(path)
        if includedInputSource is None:
            m.die(f"Tried to include a file at '{path}', but the current input type can't resolve relative paths.")
            h.removeNode(el)
            return
        doc.recordDependencies(includedInputSource)
        try:
            lines = includedInputSource.read().rawLines
        except FileNotFoundError:
            m.die(f"Couldn't find include file '{path}'.", el=el)
            h.removeNode(el)
            return
        except Exception as err:
            m.die(f"Couldn't find include file '{path}'. Error was:\n{err}", el=el)
            h.removeNode(el)
            return
        # hash the content + path together for identity
        # can't use just path, because they're relative; including "foo/bar.txt" might use "foo/bar.txt" further nested
        # can't use just content, because then you can't do the same thing twice.
        # combined does a good job unless you purposely pervert it
        hash = hashlib.md5((path + "".join(lines)).encode("ascii", "xmlcharrefreplace")).hexdigest()
        if el.get("hash"):
            hashAttr = el.get("hash")
            assert hashAttr is not None
            # This came from another included file, check if it's a loop-include
            if hash in hashAttr:
                # WHOOPS
                m.die(f"Include loop detected - “{path}” is included in itself.", el=el)
                h.removeNode(el)
                return
            hash += " " + hashAttr
        depth = int(el.get("depth") or "0")
        if depth > 100:
            # Just in case you slip past the nesting restriction
            m.die("Nesting depth > 100, literally wtf are you doing.", el=el)
            h.removeNode(el)
            return
        parseConfig = h.ParseConfig.fromSpec(doc)
        parseConfig.macros = {**parseConfig.macros, **macros}
        lines = h.parseLines(lines, parseConfig)
        lines = datablocks.transformDataBlocks(doc, lines)
        lines = markdown.parse(lines, doc.md.indent, opaqueElements=doc.md.opaqueElements)
        text = "".join(lines)
        subtree = h.parseHTML(text)
        for childInclude in h.findAll("pre.include", h.E.div({}, *subtree)):
            childInclude.set("hash", hash)
            childInclude.set("depth", str(depth + 1))
        h.replaceNode(el, *subtree)
    else:
        m.die(
            "Whoops, an include block didn't get parsed correctly, so I can't include anything.",
            el=el,
        )
        h.removeNode(el)
        return


def handleCodeInclude(el: t.ElementT, doc: t.SpecT) -> None:
    if not el.get("path"):
        m.die(
            "Whoops, an include-code block didn't get parsed correctly, so I can't include anything.",
            el=el,
        )
        h.removeNode(el)
        return
    path = el.get("path")
    assert path is not None
    includedInputSource = doc.inputSource.relative(path)
    if includedInputSource is None:
        m.die(f"Tried to include a file at '{path}', but the current input type can't resolve relative paths.")
        h.removeNode(el)
        return
    doc.recordDependencies(includedInputSource)
    try:
        lines = includedInputSource.read().rawLines
    except FileNotFoundError:
        m.die(f"Couldn't find include-code file '{path}'.", el=el)
        h.removeNode(el)
        return
    except Exception as err:
        m.die(f"Couldn't find include-code file '{path}'. Error was:\n{err}", el=el)
        h.removeNode(el)
        return
    if el.get("data-code-show"):
        codeShow = el.get("data-code-show")
        assert codeShow is not None
        showLines = parseRangeString(codeShow)
        if len(showLines) == 0:
            pass
        elif len(showLines) >= 2:
            m.die(f"Can only have one include-code 'show' segment, got '{codeShow}'.", el=el)
            return
        else:
            start, end = showLines[0]
            if end:
                lines = lines[:end]
            if start:
                lines = lines[start - 1 :]
                if not el.get("line-start"):
                    # If manually overridden, leave it alone,
                    # but otherwise DWIM.
                    el.set("line-start", str(start))
    h.appendChild(el, *lines, allowEmpty=True)


def handleRawInclude(el: t.ElementT, doc: t.SpecT) -> None:
    if not el.get("path"):
        m.die(
            "Whoops, an include-raw block didn't get parsed correctly, so I can't include anything.",
            el=el,
        )
        h.removeNode(el)
        return
    path = el.get("path", "")
    includedInputSource = doc.inputSource.relative(path)
    if includedInputSource is None:
        m.die(f"Tried to include a file at '{path}', but the current input type can't resolve relative paths.")
        h.removeNode(el)
        return
    doc.recordDependencies(includedInputSource)
    try:
        content = includedInputSource.read().content
    except FileNotFoundError:
        m.die(f"Couldn't find include-raw file '{path}'.", el=el)
        h.removeNode(el)
        return
    except Exception as err:
        m.die(f"Couldn't find include-raw file '{path}'. Error was:\n{err}", el=el)
        h.removeNode(el)
        return
    subtree = h.parseHTML(content)
    h.replaceNode(el, *subtree)


if t.TYPE_CHECKING:
    RangeItem: t.TypeAlias = list[int | None]
    # UGH I DON'T UNDERSTAND THE ERROR HERE
    # fuck it, replace this with a tiny dataclass anyway


def parseRangeString(rangeStr: str) -> list[RangeItem]:
    rangeStr = re.sub(r"\s*", "", rangeStr)
    return [_f for _f in (parseSingleRange(x) for x in rangeStr.split(",")) if _f is not None]


def parseSingleRange(item: str) -> RangeItem | None:
    if "-" in item:
        # Range, format of DDD-DDD
        low: int | None
        high: int | None
        lowStr, _, highStr = item.partition("-")
        if lowStr == "*":
            low = None
        else:
            try:
                low = int(lowStr)
            except ValueError:
                m.die(f"Error parsing include-code 'show' range '{item}' - must be `int-int`.")
                return None
        if highStr == "*":
            high = None
        else:
            try:
                high = int(highStr)
            except ValueError:
                m.die(f"Error parsing include-code 'show' range '{item}' - must be `int-int`.")
                return None
        if low is not None and high is not None and low >= high:
            m.die(f"include-code 'show' ranges must be well-formed lo-hi - got '{item}'.")
            return None
        return [low, high]
    if item == "*":
        return None
    try:
        val = int(item)
        return [val, val]
    except ValueError:
        m.die(f"Error parsing include-code 'show' value '{item}' - must be an int or *.")
        return None



from __future__ import annotations

from .stringEnum import StringEnum

dryRun: bool = False
refStatus: StringEnum = StringEnum("current", "snapshot")
biblioDisplay: StringEnum = StringEnum("index", "inline", "direct")
chroot: bool = True
executeCode: bool = False

macroStartChar = "\uebbb"
macroEndChar = "\uebbc"
incrementLineCountChar = "\uebbd"
decrementLineCountChar = "\uebbf"
bsComment = "<!--\uebbe-->"
bqStart = "\uebc0"
bqEnd = "\uebc1"



from __future__ import annotations

import dataclasses

import kdl

from .. import t
from . import utils


@dataclasses.dataclass
class Doctype:
    org: Org
    group: Group
    status: Status

    def __str__(self) -> str:
        return f"Doctype<org={self.org.name}, group={self.group.fullName()}, status={self.status.fullName()}>"


@dataclasses.dataclass
class DoctypeManager:
    genericStatuses: dict[str, Status] = dataclasses.field(default_factory=dict)
    orgs: dict[str, Org] = dataclasses.field(default_factory=dict)

    @staticmethod
    def fromKdlStr(data: str) -> DoctypeManager:
        self = DoctypeManager()
        kdlDoc = kdl.parse(data)

        for node in kdlDoc.getAll("status"):
            status = Status.fromKdlNode(node)
            self.genericStatuses[status.name] = status

        for node in kdlDoc.getAll("org"):
            org = Org.fromKdlNode(node)
            self.orgs[org.name] = org

        return self

    def getStatuses(self, name: str) -> list[Status]:
        statuses = []
        if name in self.genericStatuses:
            statuses.append(self.genericStatuses[name])
        for org in self.orgs.values():
            if name in org.statuses:
                statuses.append(org.statuses[name])
        return statuses

    def getStatus(self, orgName: str | None, statusName: str, allowGeneric: bool = False) -> Status | None:
        # Note that a None orgName does *not* indicate we don't care,
        # it's specifically statuses *not* restricted to an org.
        if orgName is None:
            return self.genericStatuses.get(statusName)
        elif orgName in self.orgs:
            statusInOrg = self.orgs[orgName].statuses.get(statusName)
            if statusInOrg:
                return statusInOrg
            elif allowGeneric:
                return self.genericStatuses.get(statusName)
            else:
                return None
        else:
            return None

    def getGroups(self, orgName: str | None, groupName: str) -> list[Group]:
        # Unlike Status, if org is None we'll just grab whatever group matches.
        groups = []
        for org in self.orgs.values():
            if orgName is not None and org.name != orgName:
                continue
            if groupName in org.groups:
                groups.append(org.groups[groupName])
        return groups

    def getGroup(self, orgName: str | None, groupName: str) -> Group | None:
        # If Org is None, and there are multiple groups with that name, fail to find.
        groups = self.getGroups(orgName, groupName)
        if len(groups) == 1:
            return groups[0]
        else:
            return None

    def getOrg(self, orgName: str) -> Org | None:
        return self.orgs.get(orgName)

    def getDoctype(self, orgName: str | None, groupName: str | None, statusName: str | None) -> Doctype:
        org, group, status = utils.canonicalize(self, orgName, groupName, statusName)
        return Doctype(org if org else NIL_ORG, group if group else NIL_GROUP, status if status else NIL_STATUS)


@dataclasses.dataclass
class Org:
    name: str
    groups: dict[str, Group] = dataclasses.field(default_factory=dict)
    statuses: dict[str, Status] = dataclasses.field(default_factory=dict)

    @staticmethod
    def fromKdlNode(node: kdl.Node) -> Org:
        name = t.cast(str, node.args[0]).upper()
        self = Org(name)
        for child in node.getAll("group"):
            g = Group.fromKdlNode(child, org=self)
            self.groups[g.name] = g
        for child in node.getAll("status"):
            s = Status.fromKdlNode(child, org=self)
            self.statuses[s.name] = s
        return self

    def __bool__(self) -> bool:
        return self != NIL_ORG


NIL_ORG = Org("(not provided)")


@dataclasses.dataclass
class Group:
    name: str
    privSec: bool
    org: Org
    requires: list[str] = dataclasses.field(default_factory=list)
    type: str | None = None

    def fullName(self) -> str:
        if self.org:
            return self.org.name + "/" + self.name
        else:
            return self.name

    @staticmethod
    def fromKdlNode(node: kdl.Node, org: Org) -> Group:
        name = t.cast(str, node.args[0]).upper()
        privSec = t.cast(bool, node.props.get("priv-sec", False))
        if "type" in node.props:
            groupType = str(node.props.get("type")).lower()
        else:
            groupType = None
        self = Group(name, privSec, org, type=groupType)
        for n in node.getAll("requires"):
            self.requires.extend(t.cast("list[str]", n.getArgs((..., str))))
        return self

    def __bool__(self) -> bool:
        return self != NIL_GROUP


NIL_GROUP = Group("(not provided)", privSec=False, org=NIL_ORG)


@dataclasses.dataclass
class Status:
    name: str
    longName: str
    org: Org
    requires: list[str] = dataclasses.field(default_factory=list)
    groupTypes: list[str] = dataclasses.field(default_factory=list)

    def fullName(self) -> str:
        if self.org:
            return self.org.name + "/" + self.name
        else:
            return self.name

    def looselyMatch(self, rawStatus: str) -> bool:
        orgName, statusName = utils.splitOrg(rawStatus)
        if statusName and self.name.upper() != statusName.upper():
            return False
        if orgName and self.org.name != orgName.upper():  # noqa: SIM103
            return False
        return True

    @staticmethod
    def fromKdlNode(node: kdl.Node, org: Org | None = None) -> Status:
        if org is None:
            org = NIL_ORG
        name = t.cast(str, node.args[0]).upper()
        longName = t.cast(str, node.args[1])
        self = Status(name, longName, org)
        for n in node.getAll("requires"):
            self.requires.extend(t.cast("list[str]", n.getArgs((..., str))))
        for n in node.getAll("group-types"):
            self.groupTypes.extend(str(x).lower() for x in n.getArgs((..., str)))
        return self

    def __bool__(self) -> bool:
        return self != NIL_STATUS


NIL_STATUS = Status("(not provided)", "", org=NIL_ORG)



# -*- coding: utf-8 -*-
# pylint: disable=arguments-renamed

from __future__ import annotations

import math as Math
from typing import TYPE_CHECKING

from . import t

if TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Dict,
        Generator,
        List,
        Tuple,
        Type,
        TypeVar,
        Union,
    )
    from typing import Optional as Opt
    from typing import Sequence as Seq

    T = TypeVar("T")
    Node: t.TypeAlias = str | "DiagramItem"  # pylint: disable=used-before-assignment
    WriterF: t.TypeAlias = Callable[[str], Any]
    WalkerF: t.TypeAlias = Callable[["DiagramItem"], Any]  # pylint: disable=used-before-assignment
    AttrsT: t.TypeAlias = Dict[str, Any]

# Display constants
DEBUG = False  # if true, writes some debug information into attributes
VS = 8  # minimum vertical separation between things. For a 3px stroke, must be at least 4
AR = 10  # radius of arcs
DIAGRAM_CLASS = "railroad-diagram"  # class to put on the root <svg>
STROKE_ODD_PIXEL_LENGTH = True  # is the stroke width an odd (1px, 3px, etc) pixel length?
INTERNAL_ALIGNMENT = "center"  # how to align items when they have extra space. left/right/center
CHAR_WIDTH = 8.5  # width of each monospace character. play until you find the right value for your font
COMMENT_CHAR_WIDTH = 7  # comments are in smaller text by default


def escapeAttr(val: Union[str, float]) -> str:
    if isinstance(val, str):
        return val.replace("&", "&amp;").replace("'", "&apos;").replace('"', "&quot;")
    return f"{val:g}"


def escapeHtml(val: str) -> str:
    return escapeAttr(val).replace("<", "&lt;")


def determineGaps(outer: float, inner: float) -> Tuple[float, float]:
    diff = outer - inner
    if INTERNAL_ALIGNMENT == "left":
        return 0, diff
    elif INTERNAL_ALIGNMENT == "right":
        return diff, 0
    else:
        return diff / 2, diff / 2


def doubleenumerate(seq: Seq[T]) -> Generator[Tuple[int, int, T], None, None]:
    length = len(list(seq))
    for i, item in enumerate(seq):
        yield i, i - length, item


def addDebug(el: DiagramItem) -> None:
    if not DEBUG:
        return
    el.attrs["data-x"] = "{0} w:{1} h:{2}/{3}/{4}".format(type(el).__name__, el.width, el.up, el.height, el.down)


class DiagramItem:
    def __init__(self, name: str, attrs: Opt[AttrsT] = None, text: Opt[Node] = None) -> None:
        self.name = name
        # up = distance it projects above the entry line
        self.up: float = 0
        # height = distance between the entry/exit lines
        self.height: float = 0
        # down = distance it projects below the exit line
        self.down: float = 0
        # width = distance between the entry/exit lines horizontally
        self.width: float = 0
        # Whether the item is okay with being snug against another item or not
        self.needsSpace = False

        # DiagramItems pull double duty as SVG elements.
        self.attrs: AttrsT = attrs or {}
        # Subclasses store their meaningful children as .item or .items;
        # .children instead stores their formatted SVG nodes.
        self.children: List[Union[Node, Path, Style]] = [text] if text else []

    def format(self, x: float, y: float, width: float) -> DiagramItem:
        raise NotImplementedError  # Virtual

    def addTo(self, parent: DiagramItem) -> DiagramItem:
        parent.children.append(self)
        return self

    def writeSvg(self, write: WriterF) -> None:
        write("<{0}".format(self.name))
        for name, value in sorted(self.attrs.items()):
            write(' {0}="{1}"'.format(name, escapeAttr(value)))
        write(">")
        if self.name in ["g", "svg"]:
            write("\n")
        for child in self.children:
            if isinstance(child, (DiagramItem, Path, Style)):
                child.writeSvg(write)
            else:
                write(escapeHtml(child))
        write("</{0}>".format(self.name))

    def walk(self, cb: WalkerF) -> None:
        cb(self)


class DiagramMultiContainer(DiagramItem):
    def __init__(
        self,
        name: str,
        items: Seq[Node],
        attrs: Opt[Dict[str, str]] = None,
        text: Opt[str] = None,
    ) -> None:
        DiagramItem.__init__(self, name, attrs, text)
        self.items: List[DiagramItem] = [wrapString(item) for item in items]

    def format(self, x: float, y: float, width: float) -> DiagramItem:
        raise NotImplementedError  # Virtual

    def walk(self, cb: WalkerF) -> None:
        cb(self)
        for item in self.items:
            item.walk(cb)


class Path:
    def __init__(self, x: float, y: float) -> None:
        self.x = x
        self.y = y
        self.attrs = {"d": f"M{x} {y}"}

    def m(self, x: float, y: float) -> Path:
        self.attrs["d"] += f"m{x} {y}"
        return self

    def l(self, x: float, y: float) -> Path:
        self.attrs["d"] += f"l{x} {y}"
        return self

    def h(self, val: float) -> Path:
        self.attrs["d"] += f"h{val}"
        return self

    def right(self, val: float) -> Path:
        return self.h(max(0, val))

    def left(self, val: float) -> Path:
        return self.h(-max(0, val))

    def v(self, val: float) -> Path:
        self.attrs["d"] += f"v{val}"
        return self

    def down(self, val: float) -> Path:
        return self.v(max(0, val))

    def up(self, val: float) -> Path:
        return self.v(-max(0, val))

    def arc_8(self, start: str, dir: str) -> Path:
        # 1/8 of a circle
        arc = AR
        s2 = 1 / Math.sqrt(2) * arc
        s2inv = arc - s2
        sweep = "1" if dir == "cw" else "0"
        path = f"a {arc} {arc} 0 0 {sweep} "
        sd = start + dir
        offset: List[float]
        if sd == "ncw":
            offset = [s2, s2inv]
        elif sd == "necw":
            offset = [s2inv, s2]
        elif sd == "ecw":
            offset = [-s2inv, s2]
        elif sd == "secw":
            offset = [-s2, s2inv]
        elif sd == "scw":
            offset = [-s2, -s2inv]
        elif sd == "swcw":
            offset = [-s2inv, -s2]
        elif sd == "wcw":
            offset = [s2inv, -s2]
        elif sd == "nwcw":
            offset = [s2, -s2inv]
        elif sd == "nccw":
            offset = [-s2, s2inv]
        elif sd == "nwccw":
            offset = [-s2inv, s2]
        elif sd == "wccw":
            offset = [s2inv, s2]
        elif sd == "swccw":
            offset = [s2, s2inv]
        elif sd == "sccw":
            offset = [s2, -s2inv]
        elif sd == "seccw":
            offset = [s2inv, -s2]
        elif sd == "eccw":
            offset = [-s2inv, -s2]
        elif sd == "neccw":
            offset = [-s2, -s2inv]

        path += " ".join(str(x) for x in offset)
        self.attrs["d"] += path
        return self

    def arc(self, sweep: str) -> Path:
        x = AR
        y = AR
        if sweep[0] == "e" or sweep[1] == "w":
            x *= -1
        if sweep[0] == "s" or sweep[1] == "n":
            y *= -1
        cw = 1 if sweep in ("ne", "es", "sw", "wn") else 0
        self.attrs["d"] += f"a{AR} {AR} 0 0 {cw} {x} {y}"
        return self

    def addTo(self, parent: DiagramItem) -> Path:
        parent.children.append(self)
        return self

    def writeSvg(self, write: WriterF) -> None:
        write("<path")
        for name, value in sorted(self.attrs.items()):
            write(f' {name}="{escapeAttr(value)}"')
        write(" />")

    def format(self) -> Path:
        self.attrs["d"] += "h.5"
        return self

    def __repr__(self) -> str:
        return f"Path({self.x!r}, {self.y!r})"


def wrapString(value: Node) -> DiagramItem:
    return value if isinstance(value, DiagramItem) else Terminal(value)


DEFAULT_STYLE = """\
    svg.railroad-diagram {
        background-color:hsl(30,20%,95%);
    }
    svg.railroad-diagram path {
        stroke-width:3;
        stroke:black;
        fill:rgba(0,0,0,0);
    }
    svg.railroad-diagram text {
        font:bold 14px monospace;
        text-anchor:middle;
    }
    svg.railroad-diagram text.label{
        text-anchor:start;
    }
    svg.railroad-diagram text.comment{
        font:italic 12px monospace;
    }
    svg.railroad-diagram rect{
        stroke-width:3;
        stroke:black;
        fill:hsl(120,100%,90%);
    }
    svg.railroad-diagram rect.group-box {
        stroke: gray;
        stroke-dasharray: 10 5;
        fill: none;
    }
"""


class Style:
    def __init__(self, css: str) -> None:
        self.css = css

    def __repr__(self) -> str:
        return f"Style({self.css!r})"

    def addTo(self, parent: DiagramItem) -> Style:
        parent.children.append(self)
        return self

    def format(self) -> Style:
        return self

    def writeSvg(self, write: WriterF) -> None:
        # Write included stylesheet as CDATA. See https://developer.mozilla.org/en-US/docs/Web/SVG/Element/style
        write(f"<style>{self.css}</style>")


class Diagram(DiagramMultiContainer):
    def __init__(self, *items: Node, **kwargs: str) -> None:
        # Accepts a type=[simple|complex] kwarg
        DiagramMultiContainer.__init__(
            self,
            "svg",
            list(items),
            {
                "class": DIAGRAM_CLASS,
            },
        )
        self.type = kwargs.get("type", "simple")
        if items and not isinstance(items[0], Start):
            self.items.insert(0, Start(self.type))
        if items and not isinstance(items[-1], End):
            self.items.append(End(self.type))
        self.up = 0
        self.down = 0
        self.height = 0
        self.width = 0
        for item in self.items:
            if isinstance(item, Style):
                continue
            self.width += item.width + (20 if item.needsSpace else 0)
            self.up = max(self.up, item.up - self.height)
            self.height += item.height
            self.down = max(self.down - item.height, item.down)
        if self.items[0].needsSpace:
            self.width -= 10
        if self.items[-1].needsSpace:
            self.width -= 10
        self.formatted = False

    def __repr__(self) -> str:
        items = ", ".join(map(repr, self.items[1:-1]))
        pieces = [] if not items else [items]
        if self.type != "simple":
            pieces.append(f"type={self.type!r}")
        return f'Diagram({", ".join(pieces)})'

    def format(
        self,
        paddingTop: float = 20,
        paddingRight: Opt[float] = None,
        paddingBottom: Opt[float] = None,
        paddingLeft: Opt[float] = None,
    ) -> Diagram:
        if paddingRight is None:
            paddingRight = paddingTop
        if paddingBottom is None:
            paddingBottom = paddingTop
        if paddingLeft is None:
            paddingLeft = paddingRight
        assert paddingRight is not None
        assert paddingBottom is not None
        assert paddingLeft is not None
        x = paddingLeft
        y = paddingTop + self.up
        g = DiagramItem("g")
        if STROKE_ODD_PIXEL_LENGTH:
            g.attrs["transform"] = "translate(.5 .5)"
        for item in self.items:
            if item.needsSpace:
                Path(x, y).h(10).addTo(g)
                x += 10
            item.format(x, y, item.width).addTo(g)
            x += item.width
            y += item.height
            if item.needsSpace:
                Path(x, y).h(10).addTo(g)
                x += 10
        self.attrs["width"] = str(self.width + paddingLeft + paddingRight)
        self.attrs["height"] = str(self.up + self.height + self.down + paddingTop + paddingBottom)
        self.attrs["viewBox"] = f"0 0 {self.attrs['width']} {self.attrs['height']}"
        g.addTo(self)
        self.formatted = True
        return self

    def writeSvg(self, write: WriterF) -> None:
        if not self.formatted:
            self.format()
        return DiagramItem.writeSvg(self, write)

    def writeStandalone(self, write: WriterF, css: str | None = None) -> None:
        if not self.formatted:
            self.format()
        if css is None:
            css = DEFAULT_STYLE
        Style(css).addTo(self)
        self.attrs["xmlns"] = "http://www.w3.org/2000/svg"
        self.attrs["xmlns:xlink"] = "http://www.w3.org/1999/xlink"
        DiagramItem.writeSvg(self, write)
        self.children.pop()
        del self.attrs["xmlns"]
        del self.attrs["xmlns:xlink"]


class Sequence(DiagramMultiContainer):
    def __init__(self, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        self.needsSpace = True
        self.up = 0
        self.down = 0
        self.height = 0
        self.width = 0
        for item in self.items:
            self.width += item.width + (20 if item.needsSpace else 0)
            self.up = max(self.up, item.up - self.height)
            self.height += item.height
            self.down = max(self.down - item.height, item.down)
        if self.items[0].needsSpace:
            self.width -= 10
        if self.items[-1].needsSpace:
            self.width -= 10
        addDebug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"Sequence({items})"

    def format(self, x: float, y: float, width: float) -> Sequence:
        leftGap, rightGap = determineGaps(width, self.width)
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).h(rightGap).addTo(self)
        x += leftGap
        for i, item in enumerate(self.items):
            if item.needsSpace and i > 0:
                Path(x, y).h(10).addTo(self)
                x += 10
            item.format(x, y, item.width).addTo(self)
            x += item.width
            y += item.height
            if item.needsSpace and i < len(self.items) - 1:
                Path(x, y).h(10).addTo(self)
                x += 10
        return self


class Stack(DiagramMultiContainer):
    def __init__(self, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        self.needsSpace = True
        self.width = max(item.width + (20 if item.needsSpace else 0) for item in self.items)
        # pretty sure that space calc is totes wrong
        if len(self.items) > 1:
            self.width += AR * 2
        self.up = self.items[0].up
        self.down = self.items[-1].down
        self.height = 0
        last = len(self.items) - 1
        for i, item in enumerate(self.items):
            self.height += item.height
            if i > 0:
                self.height += max(AR * 2, item.up + VS)
            if i < last:
                self.height += max(AR * 2, item.down + VS)
        addDebug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"Stack({items})"

    def format(self, x: float, y: float, width: float) -> Stack:
        leftGap, rightGap = determineGaps(width, self.width)
        Path(x, y).h(leftGap).addTo(self)
        x += leftGap
        xInitial = x
        if len(self.items) > 1:
            Path(x, y).h(AR).addTo(self)
            x += AR
            innerWidth = self.width - AR * 2
        else:
            innerWidth = self.width
        for i, item in enumerate(self.items):
            item.format(x, y, innerWidth).addTo(self)
            x += innerWidth
            y += item.height
            if i != len(self.items) - 1:
                (
                    Path(x, y)
                    .arc("ne")
                    .down(max(0, item.down + VS - AR * 2))
                    .arc("es")
                    .left(innerWidth)
                    .arc("nw")
                    .down(max(0, self.items[i + 1].up + VS - AR * 2))
                    .arc("ws")
                    .addTo(self)
                )
                y += max(item.down + VS, AR * 2) + max(self.items[i + 1].up + VS, AR * 2)
                x = xInitial + AR
        if len(self.items) > 1:
            Path(x, y).h(AR).addTo(self)
            x += AR
        Path(x, y).h(rightGap).addTo(self)
        return self


class OptionalSequence(DiagramMultiContainer):
    def __new__(cls, *items: Node) -> Any:
        if len(items) <= 1:
            return Sequence(*items)
        else:
            return super(OptionalSequence, cls).__new__(cls)

    def __init__(self, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        self.needsSpace = False
        self.width = 0
        self.up = 0
        self.height = sum(item.height for item in self.items)
        self.down = self.items[0].down
        heightSoFar: float = 0
        for i, item in enumerate(self.items):
            self.up = max(self.up, max(AR * 2, item.up + VS) - heightSoFar)
            heightSoFar += item.height
            if i > 0:
                self.down = (
                    max(
                        self.height + self.down,
                        heightSoFar + max(AR * 2, item.down + VS),
                    )
                    - self.height
                )
            itemWidth = item.width + (10 if item.needsSpace else 0)
            if i == 0:
                self.width += AR + max(itemWidth, AR)
            else:
                self.width += AR * 2 + max(itemWidth, AR) + AR
        addDebug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"OptionalSequence({items})"

    def format(self, x: float, y: float, width: float) -> OptionalSequence:
        leftGap, rightGap = determineGaps(width, self.width)
        Path(x, y).right(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).right(rightGap).addTo(self)
        x += leftGap
        upperLineY = y - self.up
        last = len(self.items) - 1
        for i, item in enumerate(self.items):
            itemSpace = 10 if item.needsSpace else 0
            itemWidth = item.width + itemSpace
            if i == 0:
                # Upper skip
                (
                    Path(x, y)
                    .arc("se")
                    .up(y - upperLineY - AR * 2)
                    .arc("wn")
                    .right(itemWidth - AR)
                    .arc("ne")
                    .down(y + item.height - upperLineY - AR * 2)
                    .arc("ws")
                    .addTo(self)
                )
                # Straight line
                (Path(x, y).right(itemSpace + AR).addTo(self))
                item.format(x + itemSpace + AR, y, item.width).addTo(self)
                x += itemWidth + AR
                y += item.height
            elif i < last:
                # Upper skip
                (
                    Path(x, upperLineY)
                    .right(AR * 2 + max(itemWidth, AR) + AR)
                    .arc("ne")
                    .down(y - upperLineY + item.height - AR * 2)
                    .arc("ws")
                    .addTo(self)
                )
                # Straight line
                (Path(x, y).right(AR * 2).addTo(self))
                item.format(x + AR * 2, y, item.width).addTo(self)
                (Path(x + item.width + AR * 2, y + item.height).right(itemSpace + AR).addTo(self))
                # Lower skip
                (
                    Path(x, y)
                    .arc("ne")
                    .down(item.height + max(item.down + VS, AR * 2) - AR * 2)
                    .arc("ws")
                    .right(itemWidth - AR)
                    .arc("se")
                    .up(item.down + VS - AR * 2)
                    .arc("wn")
                    .addTo(self)
                )
                x += AR * 2 + max(itemWidth, AR) + AR
                y += item.height
            else:
                # Straight line
                (Path(x, y).right(AR * 2).addTo(self))
                item.format(x + AR * 2, y, item.width).addTo(self)
                (Path(x + AR * 2 + item.width, y + item.height).right(itemSpace + AR).addTo(self))
                # Lower skip
                (
                    Path(x, y)
                    .arc("ne")
                    .down(item.height + max(item.down + VS, AR * 2) - AR * 2)
                    .arc("ws")
                    .right(itemWidth - AR)
                    .arc("se")
                    .up(item.down + VS - AR * 2)
                    .arc("wn")
                    .addTo(self)
                )
        return self


class AlternatingSequence(DiagramMultiContainer):
    def __new__(cls, *items: Node) -> AlternatingSequence:
        if len(items) == 2:
            return super(AlternatingSequence, cls).__new__(cls)
        else:
            msg = f"AlternatingSequence takes exactly two arguments, but got {len(items)} arguments."
            raise Exception(msg)

    def __init__(self, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        self.needsSpace = False

        arc = AR
        vert = VS
        first = self.items[0]
        second = self.items[1]

        arcX = 1 / Math.sqrt(2) * arc * 2
        arcY = (1 - 1 / Math.sqrt(2)) * arc * 2
        crossY = max(arc, vert)
        crossX = (crossY - arcY) + arcX

        firstOut = max(arc + arc, crossY / 2 + arc + arc, crossY / 2 + vert + first.down)
        self.up = firstOut + first.height + first.up

        secondIn = max(arc + arc, crossY / 2 + arc + arc, crossY / 2 + vert + second.up)
        self.down = secondIn + second.height + second.down

        self.height = 0

        firstWidth = (20 if first.needsSpace else 0) + first.width
        secondWidth = (20 if second.needsSpace else 0) + second.width
        self.width = 2 * arc + max(firstWidth, crossX, secondWidth) + 2 * arc
        addDebug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"AlternatingSequence({items})"

    def format(self, x: float, y: float, width: float) -> AlternatingSequence:
        arc = AR
        gaps = determineGaps(width, self.width)
        Path(x, y).right(gaps[0]).addTo(self)
        x += gaps[0]
        Path(x + self.width, y).right(gaps[1]).addTo(self)
        # bounding box
        # Path(x+gaps[0], y).up(self.up).right(self.width).down(self.up+self.down).left(self.width).up(self.down).addTo(self)
        first = self.items[0]
        second = self.items[1]

        # top
        firstIn = self.up - first.up
        firstOut = self.up - first.up - first.height
        Path(x, y).arc("se").up(firstIn - 2 * arc).arc("wn").addTo(self)
        first.format(x + 2 * arc, y - firstIn, self.width - 4 * arc).addTo(self)
        Path(x + self.width - 2 * arc, y - firstOut).arc("ne").down(firstOut - 2 * arc).arc("ws").addTo(self)

        # bottom
        secondIn = self.down - second.down - second.height
        secondOut = self.down - second.down
        Path(x, y).arc("ne").down(secondIn - 2 * arc).arc("ws").addTo(self)
        second.format(x + 2 * arc, y + secondIn, self.width - 4 * arc).addTo(self)
        Path(x + self.width - 2 * arc, y + secondOut).arc("se").up(secondOut - 2 * arc).arc("wn").addTo(self)

        # crossover
        arcX = 1 / Math.sqrt(2) * arc * 2
        arcY = (1 - 1 / Math.sqrt(2)) * arc * 2
        crossY = max(arc, VS)
        crossX = (crossY - arcY) + arcX
        crossBar = (self.width - 4 * arc - crossX) / 2
        (
            Path(x + arc, y - crossY / 2 - arc)
            .arc("ws")
            .right(crossBar)
            .arc_8("n", "cw")
            .l(crossX - arcX, crossY - arcY)
            .arc_8("sw", "ccw")
            .right(crossBar)
            .arc("ne")
            .addTo(self)
        )
        (
            Path(x + arc, y + crossY / 2 + arc)
            .arc("wn")
            .right(crossBar)
            .arc_8("s", "ccw")
            .l(crossX - arcX, -(crossY - arcY))
            .arc_8("nw", "cw")
            .right(crossBar)
            .arc("se")
            .addTo(self)
        )

        return self


class Choice(DiagramMultiContainer):
    def __init__(self, default: int, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        assert default < len(items)
        self.default = default
        self.width = AR * 4 + max(item.width for item in self.items)
        self.up = self.items[0].up
        self.down = self.items[-1].down
        self.height = self.items[default].height
        for i, item in enumerate(self.items):
            if i in [default - 1, default + 1]:
                arcs = AR * 2
            else:
                arcs = AR
            if i < default:
                self.up += max(arcs, item.height + item.down + VS + self.items[i + 1].up)
            elif i == default:
                continue
            else:
                self.down += max(
                    arcs,
                    item.up + VS + self.items[i - 1].down + self.items[i - 1].height,
                )
        self.down -= self.items[default].height  # already counted in self.height
        addDebug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return "Choice(%r, %s)" % (self.default, items)

    def format(self, x: float, y: float, width: float) -> Choice:
        leftGap, rightGap = determineGaps(width, self.width)

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).h(rightGap).addTo(self)
        x += leftGap

        innerWidth = self.width - AR * 4
        default = self.items[self.default]

        # Do the elements that curve above
        above = self.items[: self.default][::-1]
        if above:
            distanceFromY = max(AR * 2, default.up + VS + above[0].down + above[0].height)
        for i, ni, item in doubleenumerate(above):
            Path(x, y).arc("se").up(distanceFromY - AR * 2).arc("wn").addTo(self)
            item.format(x + AR * 2, y - distanceFromY, innerWidth).addTo(self)
            Path(x + AR * 2 + innerWidth, y - distanceFromY + item.height).arc("ne").down(
                distanceFromY - item.height + default.height - AR * 2,
            ).arc("ws").addTo(self)
            if ni < -1:
                distanceFromY += max(AR, item.up + VS + above[i + 1].down + above[i + 1].height)

        # Do the straight-line path.
        Path(x, y).right(AR * 2).addTo(self)
        self.items[self.default].format(x + AR * 2, y, innerWidth).addTo(self)
        Path(x + AR * 2 + innerWidth, y + self.height).right(AR * 2).addTo(self)

        # Do the elements that curve below
        below = self.items[self.default + 1 :]
        if below:
            distanceFromY = max(AR * 2, default.height + default.down + VS + below[0].up)
        for i, item in enumerate(below):
            Path(x, y).arc("ne").down(distanceFromY - AR * 2).arc("ws").addTo(self)
            item.format(x + AR * 2, y + distanceFromY, innerWidth).addTo(self)
            Path(x + AR * 2 + innerWidth, y + distanceFromY + item.height).arc("se").up(
                distanceFromY - AR * 2 + item.height - default.height,
            ).arc("wn").addTo(self)
            distanceFromY += max(
                AR,
                item.height + item.down + VS + (below[i + 1].up if i + 1 < len(below) else 0),
            )
        return self


class MultipleChoice(DiagramMultiContainer):
    def __init__(self, default: int, type: str, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        assert 0 <= default < len(items)
        assert type in ["any", "all"]
        self.default = default
        self.type = type
        self.needsSpace = True
        self.innerWidth = max(item.width for item in self.items)
        self.width = 30 + AR + self.innerWidth + AR + 20
        self.up = self.items[0].up
        self.down = self.items[-1].down
        self.height = self.items[default].height
        for i, item in enumerate(self.items):
            if i in [default - 1, default + 1]:
                minimum = 10 + AR
            else:
                minimum = AR
            if i < default:
                self.up += max(minimum, item.height + item.down + VS + self.items[i + 1].up)
            elif i == default:
                continue
            else:
                self.down += max(
                    minimum,
                    item.up + VS + self.items[i - 1].down + self.items[i - 1].height,
                )
        self.down -= self.items[default].height  # already counted in self.height
        addDebug(self)

    def __repr__(self) -> str:
        items = ", ".join(repr(item) for item in self.items)
        return f"MultipleChoice({self.default!r}, {self.type!r}, {items})"

    def format(self, x: float, y: float, width: float) -> MultipleChoice:
        leftGap, rightGap = determineGaps(width, self.width)

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).h(rightGap).addTo(self)
        x += leftGap

        default = self.items[self.default]

        # Do the elements that curve above
        above = self.items[: self.default][::-1]
        if above:
            distanceFromY = max(10 + AR, default.up + VS + above[0].down + above[0].height)
        for i, ni, item in doubleenumerate(above):
            (Path(x + 30, y).up(distanceFromY - AR).arc("wn").addTo(self))
            item.format(x + 30 + AR, y - distanceFromY, self.innerWidth).addTo(self)
            (
                Path(x + 30 + AR + self.innerWidth, y - distanceFromY + item.height)
                .arc("ne")
                .down(distanceFromY - item.height + default.height - AR - 10)
                .addTo(self)
            )
            if ni < -1:
                distanceFromY += max(AR, item.up + VS + above[i + 1].down + above[i + 1].height)

        # Do the straight-line path.
        Path(x + 30, y).right(AR).addTo(self)
        self.items[self.default].format(x + 30 + AR, y, self.innerWidth).addTo(self)
        Path(x + 30 + AR + self.innerWidth, y + self.height).right(AR).addTo(self)

        # Do the elements that curve below
        below = self.items[self.default + 1 :]
        if below:
            distanceFromY = max(10 + AR, default.height + default.down + VS + below[0].up)
        for i, item in enumerate(below):
            (Path(x + 30, y).down(distanceFromY - AR).arc("ws").addTo(self))
            item.format(x + 30 + AR, y + distanceFromY, self.innerWidth).addTo(self)
            (
                Path(x + 30 + AR + self.innerWidth, y + distanceFromY + item.height)
                .arc("se")
                .up(distanceFromY - AR + item.height - default.height - 10)
                .addTo(self)
            )
            distanceFromY += max(
                AR,
                item.height + item.down + VS + (below[i + 1].up if i + 1 < len(below) else 0),
            )
        text = DiagramItem("g", attrs={"class": "diagram-text"}).addTo(self)
        DiagramItem(
            "title",
            text=(
                "take one or more branches, once each, in any order"
                if self.type == "any"
                else "take all branches, once each, in any order"
            ),
        ).addTo(text)
        DiagramItem(
            "path",
            attrs={
                "d": "M {x} {y} h -26 a 4 4 0 0 0 -4 4 v 12 a 4 4 0 0 0 4 4 h 26 z".format(x=x + 30, y=y - 10),
                "class": "diagram-text",
            },
        ).addTo(text)
        DiagramItem(
            "text",
            text="1+" if self.type == "any" else "all",
            attrs={"x": x + 15, "y": y + 4, "class": "diagram-text"},
        ).addTo(text)
        DiagramItem(
            "path",
            attrs={
                "d": "M {x} {y} h 16 a 4 4 0 0 1 4 4 v 12 a 4 4 0 0 1 -4 4 h -16 z".format(
                    x=x + self.width - 20,
                    y=y - 10,
                ),
                "class": "diagram-text",
            },
        ).addTo(text)
        DiagramItem(
            "text",
            text="↺",
            attrs={"x": x + self.width - 10, "y": y + 4, "class": "diagram-arrow"},
        ).addTo(text)
        return self


class HorizontalChoice(DiagramMultiContainer):
    def __new__(cls, *items: Node) -> Any:
        if len(items) <= 1:
            return Sequence(*items)
        else:
            return super(HorizontalChoice, cls).__new__(cls)

    def __init__(self, *items: Node) -> None:
        DiagramMultiContainer.__init__(self, "g", items)
        allButLast = self.items[:-1]
        middles = self.items[1:-1]
        first = self.items[0]
        last = self.items[-1]
        self.needsSpace = False

        self.width = (
            AR  # starting track
            + AR * 2 * (len(self.items) - 1)  # inbetween tracks
            + sum(x.width + (20 if x.needsSpace else 0) for x in self.items)  # items
            + (AR if last.height > 0 else 0)  # needs space to curve up
            + AR
        )  # ending track

        # Always exits at entrance height
        self.height = 0

        # All but the last have a track running above them
        self._upperTrack = max(AR * 2, VS, max(x.up for x in allButLast) + VS)
        self.up = max(self._upperTrack, last.up)

        # All but the first have a track running below them
        # Last either straight-lines or curves up, so has different calculation
        self._lowerTrack = max(
            VS,
            max(x.height + max(x.down + VS, AR * 2) for x in middles) if middles else 0,
            last.height + last.down + VS,
        )
        if first.height < self._lowerTrack:
            # Make sure there's at least 2*AR room between first exit and lower track
            self._lowerTrack = max(self._lowerTrack, first.height + AR * 2)
        self.down = max(self._lowerTrack, first.height + first.down)

        addDebug(self)

    def format(self, x: float, y: float, width: float) -> HorizontalChoice:
        # Hook up the two sides if self is narrower than its stated width.
        leftGap, rightGap = determineGaps(width, self.width)
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).h(rightGap).addTo(self)
        x += leftGap

        first = self.items[0]
        last = self.items[-1]

        # upper track
        upperSpan = (
            sum(x.width + (20 if x.needsSpace else 0) for x in self.items[:-1]) + (len(self.items) - 2) * AR * 2 - AR
        )
        (Path(x, y).arc("se").up(self._upperTrack - AR * 2).arc("wn").h(upperSpan).addTo(self))

        # lower track
        lowerSpan = (
            sum(x.width + (20 if x.needsSpace else 0) for x in self.items[1:])
            + (len(self.items) - 2) * AR * 2
            + (AR if last.height > 0 else 0)
            - AR
        )
        lowerStart = x + AR + first.width + (20 if first.needsSpace else 0) + AR * 2
        (
            Path(lowerStart, y + self._lowerTrack)
            .h(lowerSpan)
            .arc("se")
            .up(self._lowerTrack - AR * 2)
            .arc("wn")
            .addTo(self)
        )

        # Items
        for [i, item] in enumerate(self.items):
            # input track
            if i == 0:
                (Path(x, y).h(AR).addTo(self))
                x += AR
            else:
                (Path(x, y - self._upperTrack).arc("ne").v(self._upperTrack - AR * 2).arc("ws").addTo(self))
                x += AR * 2

            # item
            itemWidth = item.width + (20 if item.needsSpace else 0)
            item.format(x, y, itemWidth).addTo(self)
            x += itemWidth

            # output track
            if i == len(self.items) - 1:
                if item.height == 0:
                    (Path(x, y).h(AR).addTo(self))
                else:
                    (Path(x, y + item.height).arc("se").addTo(self))
            elif i == 0 and item.height > self._lowerTrack:
                # Needs to arc up to meet the lower track, not down.
                if item.height - self._lowerTrack >= AR * 2:
                    (
                        Path(x, y + item.height)
                        .arc("se")
                        .v(self._lowerTrack - item.height + AR * 2)
                        .arc("wn")
                        .addTo(self)
                    )
                else:
                    # Not enough space to fit two arcs
                    # so just bail and draw a straight line for now.
                    (Path(x, y + item.height).l(AR * 2, self._lowerTrack - item.height).addTo(self))
            else:
                (Path(x, y + item.height).arc("ne").v(self._lowerTrack - item.height - AR * 2).arc("ws").addTo(self))
        return self


def Optional(item: Node, skip: bool = False) -> Choice:
    return Choice(0 if skip else 1, Skip(), item)


class OneOrMore(DiagramItem):
    def __init__(self, item: Node, repeat: Opt[Node] = None) -> None:
        DiagramItem.__init__(self, "g")
        self.item = wrapString(item)
        repeat = repeat or Skip()
        self.rep = wrapString(repeat)
        self.width = max(self.item.width, self.rep.width) + AR * 2
        self.height = self.item.height
        self.up = self.item.up
        self.down = max(AR * 2, self.item.down + VS + self.rep.up + self.rep.height + self.rep.down)
        self.needsSpace = True
        addDebug(self)

    def format(self, x: float, y: float, width: float) -> OneOrMore:
        leftGap, rightGap = determineGaps(width, self.width)

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).h(rightGap).addTo(self)
        x += leftGap

        # Draw item
        Path(x, y).right(AR).addTo(self)
        self.item.format(x + AR, y, self.width - AR * 2).addTo(self)
        Path(x + self.width - AR, y + self.height).right(AR).addTo(self)

        # Draw repeat arc
        distanceFromY = max(AR * 2, self.item.height + self.item.down + VS + self.rep.up)
        Path(x + AR, y).arc("nw").down(distanceFromY - AR * 2).arc("ws").addTo(self)
        self.rep.format(x + AR, y + distanceFromY, self.width - AR * 2).addTo(self)
        Path(x + self.width - AR, y + distanceFromY + self.rep.height).arc("se").up(
            distanceFromY - AR * 2 + self.rep.height - self.item.height,
        ).arc("en").addTo(self)

        return self

    def walk(self, cb: WalkerF) -> None:
        cb(self)
        self.item.walk(cb)
        self.rep.walk(cb)

    def __repr__(self) -> str:
        return f"OneOrMore({self.item!r}, repeat={self.rep!r})"


def ZeroOrMore(item: Node, repeat: Opt[Node] = None, skip: bool = False) -> Choice:
    result = Optional(OneOrMore(item, repeat), skip)
    return result


class Group(DiagramItem):
    def __init__(self, item: Node, label: Opt[Node] = None) -> None:
        DiagramItem.__init__(self, "g")
        self.item = wrapString(item)
        self.label: Opt[DiagramItem]
        if isinstance(label, DiagramItem):
            self.label = label
        elif label:
            self.label = Comment(label)
        else:
            self.label = None

        self.width = max(
            self.item.width + (20 if self.item.needsSpace else 0),
            self.label.width if self.label else 0,
            AR * 2,
        )
        self.height = self.item.height
        self.boxUp = max(self.item.up + VS, AR)
        self.up = self.boxUp
        if self.label:
            self.up += self.label.up + self.label.height + self.label.down
        self.down = max(self.item.down + VS, AR)
        self.needsSpace = True
        addDebug(self)

    def format(self, x: float, y: float, width: float) -> Group:
        leftGap, rightGap = determineGaps(width, self.width)
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y + self.height).h(rightGap).addTo(self)
        x += leftGap

        DiagramItem(
            "rect",
            {
                "x": x,
                "y": y - self.boxUp,
                "width": self.width,
                "height": self.boxUp + self.height + self.down,
                "rx": AR,
                "ry": AR,
                "class": "group-box",
            },
        ).addTo(self)

        self.item.format(x, y, self.width).addTo(self)
        if self.label:
            self.label.format(
                x,
                y - (self.boxUp + self.label.down + self.label.height),
                self.label.width,
            ).addTo(self)

        return self

    def walk(self, cb: WalkerF) -> None:
        cb(self)
        self.item.walk(cb)
        if self.label:
            self.label.walk(cb)


class Start(DiagramItem):
    def __init__(self, type: str = "simple", label: Opt[str] = None) -> None:
        DiagramItem.__init__(self, "g")
        if label:
            self.width = max(20, len(label) * CHAR_WIDTH + 10)
        else:
            self.width = 20
        self.up = 10
        self.down = 10
        self.type = type
        self.label = label
        addDebug(self)

    def format(self, x: float, y: float, width: float) -> Start:
        path = Path(x, y - 10)
        if self.type == "complex":
            path.down(20).m(0, -10).right(self.width).addTo(self)
        else:
            path.down(20).m(10, -20).down(20).m(-10, -10).right(self.width).addTo(self)
        if self.label:
            DiagramItem(
                "text",
                attrs={"x": x, "y": y - 15, "style": "text-anchor:start"},
                text=self.label,
            ).addTo(self)
        return self

    def __repr__(self) -> str:
        return f"Start(type={self.type!r}, label={self.label!r})"


class End(DiagramItem):
    def __init__(self, type: str = "simple") -> None:
        DiagramItem.__init__(self, "path")
        self.width = 20
        self.up = 10
        self.down = 10
        self.type = type
        addDebug(self)

    def format(self, x: float, y: float, width: float) -> End:
        if self.type == "simple":
            self.attrs["d"] = "M {0} {1} h 20 m -10 -10 v 20 m 10 -20 v 20".format(x, y)
        elif self.type == "complex":
            self.attrs["d"] = "M {0} {1} h 20 m 0 -10 v 20".format(x, y)
        return self

    def __repr__(self) -> str:
        return f"End(type={self.type!r})"


class Terminal(DiagramItem):
    def __init__(self, text: str, href: Opt[str] = None, title: Opt[str] = None, cls: str = "") -> None:
        DiagramItem.__init__(self, "g", {"class": " ".join(["terminal", cls])})
        self.text = text
        self.href = href
        self.title = title
        self.cls = cls
        self.width = len(text) * CHAR_WIDTH + 20
        self.up = 11
        self.down = 11
        self.needsSpace = True
        addDebug(self)

    def __repr__(self) -> str:
        return f"Terminal({self.text!r}, href={self.href!r}, title={self.title!r}, cls={self.cls!r})"

    def format(self, x: float, y: float, width: float) -> Terminal:
        leftGap, rightGap = determineGaps(width, self.width)

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y).h(rightGap).addTo(self)

        DiagramItem(
            "rect",
            {
                "x": x + leftGap,
                "y": y - 11,
                "width": self.width,
                "height": self.up + self.down,
                "rx": 10,
                "ry": 10,
            },
        ).addTo(self)
        text = DiagramItem("text", {"x": x + leftGap + self.width / 2, "y": y + 4}, self.text)
        if self.href is not None:
            a = DiagramItem("a", {"xlink:href": self.href}, text).addTo(self)
            text.addTo(a)
        else:
            text.addTo(self)
        if self.title is not None:
            DiagramItem("title", {}, self.title).addTo(self)
        return self


class NonTerminal(DiagramItem):
    def __init__(self, text: str, href: Opt[str] = None, title: Opt[str] = None, cls: str = "") -> None:
        DiagramItem.__init__(self, "g", {"class": " ".join(["non-terminal", cls])})
        self.text = text
        self.href = href
        self.title = title
        self.cls = cls
        self.width = len(text) * CHAR_WIDTH + 20
        self.up = 11
        self.down = 11
        self.needsSpace = True
        addDebug(self)

    def __repr__(self) -> str:
        return f"NonTerminal({self.text!r}, href={self.href!r}, title={self.title!r}, cls={self.cls!r})"

    def format(self, x: float, y: float, width: float) -> NonTerminal:
        leftGap, rightGap = determineGaps(width, self.width)

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y).h(rightGap).addTo(self)

        DiagramItem(
            "rect",
            {
                "x": x + leftGap,
                "y": y - 11,
                "width": self.width,
                "height": self.up + self.down,
            },
        ).addTo(self)
        text = DiagramItem("text", {"x": x + leftGap + self.width / 2, "y": y + 4}, self.text)
        if self.href is not None:
            a = DiagramItem("a", {"xlink:href": self.href}, text).addTo(self)
            text.addTo(a)
        else:
            text.addTo(self)
        if self.title is not None:
            DiagramItem("title", {}, self.title).addTo(self)
        return self


class Comment(DiagramItem):
    def __init__(self, text: str, href: Opt[str] = None, title: Opt[str] = None, cls: str = "") -> None:
        DiagramItem.__init__(self, "g", {"class": " ".join(["non-terminal", cls])})
        self.text = text
        self.href = href
        self.title = title
        self.cls = cls
        self.width = len(text) * COMMENT_CHAR_WIDTH + 10
        self.up = 8
        self.down = 8
        self.needsSpace = True
        addDebug(self)

    def __repr__(self) -> str:
        return f"Comment({self.text!r}, href={self.href!r}, title={self.title!r}, cls={self.cls!r})"

    def format(self, x: float, y: float, width: float) -> Comment:
        leftGap, rightGap = determineGaps(width, self.width)

        # Hook up the two sides if self is narrower than its stated width.
        Path(x, y).h(leftGap).addTo(self)
        Path(x + leftGap + self.width, y).h(rightGap).addTo(self)

        text = DiagramItem(
            "text",
            {"x": x + leftGap + self.width / 2, "y": y + 5, "class": "comment"},
            self.text,
        )
        if self.href is not None:
            a = DiagramItem("a", {"xlink:href": self.href}, text).addTo(self)
            text.addTo(a)
        else:
            text.addTo(self)
        if self.title is not None:
            DiagramItem("title", {}, self.title).addTo(self)
        return self


class Skip(DiagramItem):
    def __init__(self) -> None:
        DiagramItem.__init__(self, "g")
        self.width = 0
        self.up = 0
        self.down = 0
        addDebug(self)

    def format(self, x: float, y: float, width: float) -> Skip:
        Path(x, y).right(width).addTo(self)
        return self

    def __repr__(self) -> str:
        return "Skip()"



from __future__ import annotations

import dataclasses
import os

from . import h, t


@dataclasses.dataclass
class OutlineEntry:
    text: str
    id: str | None
    level: int
    lineNum: int


def printOutline(doc: t.Spec) -> str:
    entries = generateOutline(doc)
    lineNumWidth = max(len(str(e.lineNum)) for e in entries)
    idWidth = max(len(str(e.id or "")) for e in entries)
    consoleWidth = os.get_terminal_size().columns
    lines = []
    for entry in entries:
        if entry.level == 2:
            prefix = ""
        else:
            prefix = " " * (entry.level - 2)
        line = f"{entry.lineNum:{lineNumWidth}} | #{entry.id:{idWidth}} | {prefix}{entry.text}"
        if len(line) > consoleWidth:
            line = line[: consoleWidth - 2] + "…"
        lines.append(line)
    return "\n".join(lines)


def generateOutline(doc: t.Spec) -> list[OutlineEntry]:
    entries = []
    for el in collectHeadings(doc):
        text = h.textContent(el).strip().replace("\n", "")
        id = el.get("id", None)
        level = int((h.tagName(el) or "h0")[1])
        try:
            lineNum = int(el.get("bs-line-number", ""))
        except:
            try:
                num, _, rest = el.get("bs-line-number", "").partition(":")
                lineNum = int(num)
                int(rest)
            except:  # noqa: S112
                continue
        entries.append(OutlineEntry(text, id, level, lineNum))
    return entries


def collectHeadings(
    doc: t.Spec,
    root: t.ElementT | None = None,
    headings: list[t.ElementT] | None = None,
) -> list[t.ElementT]:
    if headings is None:
        headings = []
    if root is None:
        root = doc.body
    for child in h.childElements(root):
        if h.tagName(child) in ("h1", "h2", "h3", "h4", "h5", "h6"):
            headings.append(child)
        collectHeadings(doc, child, headings)
    return headings



from . import parser
from .dom import (
    DuplicatedLinkText,
    E,
    addClass,
    addOldIDs,
    appendChild,
    appendContents,
    approximateLineNumber,
    childElements,
    childNodes,
    circledDigits,
    clearContents,
    closestAncestor,
    closestAttr,
    collectAutolinks,
    collectIds,
    collectLinksWithSectionNames,
    collectSyntaxHighlightables,
    createElement,
    dedupIDs,
    emptyText,
    escapeAttr,
    escapeCSSIdent,
    escapeHTML,
    escapeUrlFrag,
    filterAncestors,
    find,
    findAll,
    firstLinkTextFromElement,
    fixSurroundingTypography,
    fixupIDs,
    foldWhitespace,
    hasAncestor,
    hasAnyAttr,
    hasAttr,
    hasAttrs,
    hasChildElements,
    hasClass,
    hashContents,
    hasOnlyChild,
    headingLevelOfElement,
    innerHTML,
    insertAfter,
    insertBefore,
    isElement,
    isEmpty,
    isNormative,
    isOddNode,
    isOnlyChild,
    linkTextsFromElement,
    moveContents,
    nextSiblingElement,
    nextSiblingNode,
    nodeIter,
    outerHTML,
    parentElement,
    parseDocument,
    parseHTML,
    prependChild,
    previousElements,
    printNodeTree,
    relevantHeadings,
    removeAttr,
    removeClass,
    removeNode,
    renameAttr,
    replaceContents,
    replaceMacrosTextly,
    replaceNode,
    replaceWithContents,
    safeID,
    scopingElements,
    sectionName,
    serializeTag,
    sortElements,
    tagName,
    textContent,
    textContentIgnoringDecorative,
    transferAttributes,
    treeAttr,
    unescape,
    unfixTypography,
    uniqueID,
    wrapContents,
)
from .parser import (
    ParseConfig,
    debugNodes,
    initialDocumentParse,
    nodesFromHtml,
    parseLines,
    parseText,
    parseTitle,
    strFromNodes,
)
from .serializer import Serializer
